;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; manumission.scm - source code for manumission HTML Helpfile builder
;;;
;;; Copyright (C) 2009 Caliber Technology LLC
;;;
;;; This file is part of manumission
;;;
;;; manumission is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; manumission is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with manumission. If not, see <http://www.gnu.org/licenses/>.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; A very general diagram of the chm format:
;;;
;;; --------------------
;;; | header           |
;;; --------------------
;;; | directory        |
;;; --------------------
;;; | content sections |
;;; -------------------
;;;
;;; The directory, also referred to as a listing chunk in the
;;; specs, amounts to a alphabetically-sorted list of all the
;;; internal virtual "files" found in the content section.
;;;
;;; There are two content sections.  The first (content section 0)
;;; is uncompressed, meaning all the virtual files within it can
;;; be read directly.  The second (content section 1) is compressed
;;; using lzx compression.  For now, we're just putting all the
;;; "files" in the uncompressed section.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; This is a command-line tool, and the goal here is to minimize
;;; configuration duties.  We load the 'in-dir' files as pages,
;;; and look for the special names "contents.hhc" and "index.hhk"
;;; to determine the contents and index pages.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(module manumission
  (main main))

(define *program* "manumission")
(define *version* "0.1")

;; These globals are place-holders for work that's pending.  Each
;; will need to either be computed or read from config info.

(define *compatibility* 1.1)       ; 1.1 or later
(define *index-tree-depth* 1)      ; depth of index tree (1=no index, 2=one level of PMGI chunks
(define *quickref-density* 2)      ; normally 2
(define *directory-chunk-size* #x1000)
(define *windows-language-id* #x409)
(define *HHP-language-id* #x409)
(define *hhp-default-font* "Ariel,12,0")
(define *hhp-compiled-file* "test")  ; file base name -> lowercase, without extension
(define *hhp-timestamp* 0)
(define *file-has-Klinks?* 0)
(define *file-has-Alinks?* 0)
(define *full-text-search-on?* 0)
(define *DBCS-in-use?* 0)
(define *binary-index/TOC-dword* #x1234)
(define *number-information-types* 0)
(define *nav-pane-width* 150)            ; width of the navigation pane in pixels.
(define *text/site-ImageList-offset* 0)  ; offset in /#STRINGS of the ImageList param of the "text/site 
;;                                         properties" object of the sitemap contents
(define *text/site-ImageList-folder?* 0)
(define *text/site-Background* 0)
(define *text/site-Foreground* 0)
(define *text/site-Font* 0)
(define *text/site-Window-Styles* 0)
(define *text/site-ExWindow-Styles* 0)
(define *text/site-FrameName* 0)
(define *text/site-WindowName* 0)
(define *number-of-MERGE-FILES* 0)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; remove any duplicate items
(define (keep-unique lst)
  (let loop ((lst lst)
	     (acc '()))
    (if (null? lst)
	(reverse acc)
	(let ((item (car lst)))
	  (loop (cdr lst) 
		(if (member item (cdr lst))
		    acc
		    (cons item acc)))))))

;; remove elements unless test is #t
(define (keep-not test lst)
  (filter
   (lambda (e)
     (not (test e)))
   lst))

;; compose 2 functions:
(define (compose f g)
  (lambda args
    (f (apply g args))))

(define-record-type topic
  (make-topic name section data attributes)
  topic?
  (name topic-name set-topic-name!)
  (section topic-section set-topic-section!)
  (data topic-data set-topic-data!)
  (attributes topic-attributes set-topic-attributes))

;; loads html files into a list of
;; '(path content-string . other)
;; currently the optional third element is used
;; to indicate special pages
(define (load-dir dir default-topic contents-topic index-topic)
  (let ((names (directory->list dir)))
    (map (lambda (p)
	   (make-topic
	    (format "/~a" p)
	    0   ; section 0 for now
	    (file->string (format "~a/~a" dir p))
	    (cond
	     ((string=? p default-topic) 'default)
	     ((string=? p contents-topic) 'contents)
	     ((string=? p index-topic) 'index)
	     (else #f))))
	 names)))

;; a special topic, like contents or index?
(define (special? topic)
  (memq (topic-attributes topic) '(contents index)))

;; return the filename for a special topic
(define (find-topic-name sym topics)
  (let* ((a-list (filter-map
		  (lambda (topic)
		    (if (special? topic)
			(cons (topic-attributes topic) topic)
			#f))
		  topics))
	 (fnd (assq sym a-list)))
    (if fnd (ignore-/ (topic-name (cdr fnd))) "")))
    
;; drop initial "/" in path
(define (ignore-/ path)
  (if (and (not (string=? path ""))
	   (eq? #\/ (string-ref path 0)))
      (substring path 1)
      path))

;; return the title of html page
(define (page-title str)
  (let* ((p1 "(?:T|t)(?:I|i)(?:T|t)(?:L|l)(?:E|e)")
	 (open (format "<~a>" p1))
	 (close (format "</~a>" p1))
	 (m (pregexp-match (format "~a(.*)~a" open close) str)))
    (if m
	(cadr m)
	#f)))

(define (bit-or* . args)
  (let loop ((lst args)
	     (acc 0))
    (if (null? lst)
	acc
	(loop (cdr lst)
	      (bit-or (car lst) acc)))))

;; filters out #f items in lst
(define (remove-false lst)
  (let loop ((lst lst)
	     (acc '()))
    (if (null? lst)
	(reverse acc)
	(loop (cdr lst)
	      (if (car lst)
		  (cons (car lst) acc)
		  acc)))))

;; converts from string to utf-8
(define (utf-8 str)
  str)

;; converts from string to utf-16.
;; (we just insert null bytes)
(define (utf-16 str)
  (let loop ((lst (string->list str))
	     (acc '()))
    (if (null? lst)
	(list->string (reverse acc))
	(loop (cdr lst)
	      (append (list #\null (car lst)) acc)))))

;; helper for integer output
(define (int->bytes val len)
  (list->string
   (map
    integer->char
    (let loop ((n val)
	       (acc '()))
      (if (>= (length acc) len)
	  (reverse acc)
	  (loop (quotient n 256)
		(cons (remainder n 256) acc)))))))

;; returns val as a QWORD byte string
(define (qword val)
  (int->bytes val 8))

;; returns val as a DWORD byte string
(define (dword val)
  (int->bytes val 4))

;; returns val as a WORD byte string
(define (word val)
  (int->bytes val 2))

;; returns the which'th byte from val
(define (get-byte val which)
  (make-string
   1
   (integer->char
    (bit-rsh
     (bit-and
      (bit-lsh #xFF (* 8 which)) val) (* 8 which)))))

;; returns a WORD with bytes reversed
(define (word-as-2-bytes val)
  (string-append
   (get-byte val 1)
   (get-byte val 0)))

;; total number of bytes in a list of byte strings
(define (byte-list-total blist)
  (string-length (apply string-append blist)))

;; returns the byte-string for an ENCINT
(define (encint val)
  (let loop ((v (bit-rsh val 7))
	     (acc (list (make-string 1 (integer->char (bit-and #x7F val))))))
    (if (zero? v)
	(apply string-append acc)
	(loop (bit-rsh v 7)
	      (cons (make-string 1 (integer->char (bit-or #x80 (bit-and #x7F v)))) acc)))))

;; returns the guid as a byte-string
(define (guid str)
  (let* ((val (if (eq? #\{ (string-ref str 0))
		  (substring str 1 (- (string-length str) 1))))
	 (parts (map (lambda (s) (string->number s 16))
		     (string-split val "-"))))
    (string-append
     (dword (car parts))
     (word (list-ref parts 1))
     (word (list-ref parts 2))
     (word-as-2-bytes (list-ref parts 3))
     (word-as-2-bytes (list-ref parts 4))
     (word-as-2-bytes (list-ref parts 5))
     (word-as-2-bytes (list-ref parts 6)))))

;; is the topic in the TOC?
(define (in-contents? topic)
  #t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Now we're getting serious
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; first bytes in the chm file
(define (prolog)
  (string-append
   "ITSF"
   (dword 3)      ; version
   (dword 96)     ; header length
   (dword 1)
   (dword 2862673702)    ; timestamp
   (dword #x409)  ; Windows language id
   (guid "{7C01FD10-7BAA-11D0-9E0C-00A0-C922-E6EC}")
   (guid "{7C01FD11-7BAA-11D0-9E0C-00A0-C922-E6EC}")))

;; the header-section-table
(define (header-section-table prolog headers)
  (apply string-append
	 (let loop ((lst (reverse headers))
		    (acc '()))
	   (if (null? lst)
	       acc
	       (loop (cdr lst)
		     (cons
		      (string-append
		       ;; section header offset from file start
		       (qword (+ (string-length prolog)
				 ;; section table length
				 (* (length headers) #x10) 8
				 ;; length of preceding header sections:
				 (if (= 1 (length lst)) 0
				     (string-length (apply string-append (cdr lst))))))
		       ;; section header table - length
		       (qword (string-length (car lst))))
		      acc))))))

;; header section 0
(define (header-section0 filesize)
  (apply string-append
	 (list
	  (dword #x01FE)
	  (dword 0)
	  (qword filesize)
	  (dword 0)
	  (dword 0))))

;; header section 1
(define (header-section1 index-chunks dir-listings)
  (string-append
   (apply string-append
	  (list
	   "ITSP"
	   (dword 1)       ; version
	   (dword #x54)    ; length of directory header
	   (dword #x0a)    ; unknown
	   (dword *directory-chunk-size*)
	   (dword *quickref-density*)
	   (dword (if (null? index-chunks) 1 2))
	   (dword (if (null? index-chunks) #xFFFFFFFF 0))
	   (dword (length index-chunks))
	   (dword (+ (length index-chunks)
		     (length dir-listings)
		     -1))
	   (dword #xFFFFFFFF)
	   (dword (length dir-listings))
	   (dword *windows-language-id*)
	   (guid "{5D02926A-212E-11D0-9DF9-00A0-C922-E6EC}")
	   (dword #x54)    ; length of directory header
	   (dword #xFFFFFFFF)
	   (dword #xFFFFFFFF)
	   (dword #xFFFFFFFF)))
   (apply
    string-append 
    (append
     index-chunks
     (map car dir-listings)))))

;; returns a single directory entry
;; name is a string
(define (make-entry name section-num offset len)
  ;;(printf "make-entry ~s ~s ~s ~s\n" name section-num offset len)
  (string-append
   (encint (string-length name))
   (utf-8 name)
   (encint section-num)
   (encint offset)
   (encint len)))

;; returns a list of quickref entries
;; 1 entry for every fifth file
;; each is an offset from entry 0
(define (quickref e-list)
  (let ((modulus (+ 1 (bit-lsh *quickref-density* 1))))
    (let loop ((lst e-list)
	       (count 0)
	       (offset 0)
	       (acc (list (length e-list))))    ;; last entry is number of entries
      (cond
       ((null? lst) acc)
       ((and (not (zero? count))
	     (zero? (modulo count modulus)))
	(loop (cdr lst) (+ 1 count)
	      (+ offset (string-length (car lst)))
	      (cons offset acc)))
       (else (loop (cdr lst)
		   (+ 1 count)
		   (+ offset (string-length (car lst)))
		   acc))))))

;; directory listing-chunks
;; f-list is ((name content-section-number offset length) ...)
;; which is filename, content-section-number, offset in content
;; section.
;; returns a list of
;; (list chunk first-name)
;; where chunk is the listing-chunk string, 
;; and first-name is the name of it's first entry
;; (used when building an index chunk).
(define (listing-chunks f-list)
  (let loop ((lst f-list)
	     (acc '()))
    (if (null? lst)
	(reverse acc)
	(let* ((chunk (make-listing-chunk lst))
	       (remaining (drop lst (cadr chunk))))
	  (loop remaining (cons chunk acc))))))

;; bytes used by current quickref area:
(define (quick-space entries)
  (* 4
     (+ 1
	(quotient
	 (length entries)
	 (+ 1 (bit-lsh #x1 *quickref-density*))))))

;; takes the most entries it can from f-list,
;; makes a listing-chunk, and returns a list of
;; (list chunk count first-name)
;; where count is the number of names used
;; and first-name the first entry in the chunk
(define (make-listing-chunk f-list)
  ;;(printf "f-list:\n") (pp f-list)
  (let loop ((lst f-list)
	     (count 0)
	     (offset 0)
	     (first-name #f)
	     (entries '()))
    (if (null? lst)
	(let* ((quicks (map word (quickref entries)))
	       (free-space (- *directory-chunk-size*
			      #x14   ; offset of first entry in the chunk
			      (byte-list-total entries)))
	       (padding-len (- free-space (byte-list-total quicks))))
	  (list
	   (string-append
	    "PMGL"
	    (dword free-space)  ; Length of free space and/or quickref area at end of directory chunk
	    (dword 0)
	    (dword #xFFFFFFFF)  ; chunk number of previous listing-chunk!!!
	    (dword #xFFFFFFFF)  ; chunk number of next listing-chunk!!!
	    (apply string-append (reverse entries))
	    (make-string padding-len)
	    (apply string-append quicks))
	   count
	   (ignore-/ first-name)))
	(let* ((item (car lst))
	       (entry (make-entry (car item) (cadr item) (cadddr item) (list-ref item 4)))
	       (name (ignore-/ (caar lst)))
	       (free-space (- *directory-chunk-size*
			      #x14   ; offset of first entry in the chunk
			      (byte-list-total entries))))
	  (cond
	   ((positive? (- free-space
			  (string-length entry)
			  (quick-space entries)))
	    (loop (cdr lst)
		  (+ count 1)
		  (+ offset (string-length entry))
		  (if first-name first-name  name)
		  (cons entry entries)))
	   (else (loop '() count offset first-name entries)))))))

;; start of file
(define (header index-chunk-list dir-listings filesize)
    (let* ((p0 (prolog))
	   (h0 (header-section0 filesize))
	   (h1 (header-section1 index-chunk-list dir-listings))
	   (tbl (header-section-table p0 (list h0 h1)))
	   (h-addl (qword (+ 8
			     (string-length p0)
			     (string-length tbl)
			     (string-length h0)
			     (string-length h1)))))
      (string-append p0 tbl h-addl h0 h1)))

;; returns the #SYSTEM coded entry
;; if add-NT? is #t, add a null-terminator
(define (system$entry code data . add-NT?)
  (let* ((add-nt? (if (null? add-NT?) #f (car add-NT?)))
	 (entry (if add-nt? (string-append data (make-string 1 #\000)) data)))
    (string-append
     (word code) (word (string-length entry)) entry)))

;; creates the #SYSTEM "file"
(define (system$ topics title)
  (string-append
   (dword (if (= *compatibility* 1) 2 3)) ; version
   (system$entry 10 (dword #x27192300)); timestamp
   (system$entry  9 *program* #t)
   (system$entry  4 (string-append (dword *HHP-language-id*)
				   (dword *DBCS-in-use?*)
				   (dword *full-text-search-on?*)
				   (dword *file-has-Klinks?*)
				   (dword *file-has-Alinks?*)
				   (qword *hhp-timestamp*)
				   (dword 0)       ; unknown
				   (dword 0)))     ; unknown
   (system$entry  2 (find-topic-name 'default topics) #t)
   (system$entry  3 title #t)
   (system$entry 16 *hhp-default-font* #t)
   (system$entry  6 *hhp-compiled-file* #t)
   (system$entry  0 (find-topic-name 'contents topics) #t)
   (system$entry  1 (find-topic-name 'index topics) #t)
   (system$entry  7 (dword *binary-index/TOC-dword*))   ; when Binary Index on, must match dword at /#URLTBL->sitemap
   (system$entry 11 (dword *binary-index/TOC-dword*))   ; when Binary TOC on, must match dword at /#URLTBL->sitemap
   (system$entry 12 (dword *number-information-types*))
   ;; skip code 13
   ;; skip code 14
   ;; skip code 8
   (if (= *compatibility* 1)
       ""
       (system$entry 15 (dword 0)))  ; Information type checksum. Unknown algorithm & data source.
   ))

;; the #IDXHDR "file"
(define (idxhdr$ topics)
  (string-append
   "T#SM"
   (dword 8653545)   ; offset  4 unknown (timestamp/checksum ?)
   (dword 1)   ; offset  8 unknown
   (dword (length topics)) ; offset  C
   (dword 0)   ; offset 10 unknown
   (dword *text/site-ImageList-offset*)    ; offset 14
   (dword 0)   ; offset 18 unknown
   (dword *text/site-ImageList-folder?*)   ; offset 1C
   (dword *text/site-Background*)          ; offset 20
   (dword *text/site-Foreground*)          ; offset 24
   (dword *text/site-Font*)                ; offset 28
   (dword *text/site-Window-Styles*)       ; offset 2C
   (dword *text/site-FrameName*)           ; offset 38
   (dword *text/site-ExWindow-Styles*)     ; offset 30
   (dword 0) ; offset 34 Unknown. Often -1. Sometimes 0.
   (dword *text/site-WindowName*)          ; offset 3C
   (dword *number-information-types*)      ; offset 40
   (dword 1) ; offset 44 Unknown.  Often 1. Also 0, 3.
   (dword *number-of-MERGE-FILES*)         ; offset 48
   (dword 0) ; offset 4C Unknown.  Often 0. Non-zero mostly in files with some files in the merge files list.
   (make-string 4016 #\000); offset 50 List of offsets in the #STRINGS file that are the [MERGE FILES] list
   ))

;; the #STRINGS "file"
;; string-list is a list of strings for the file
(define (strings$ string-list)
  (apply string-append 
	 (cons "\0"
	       (map (lambda (s)
		      (string-append
		       (utf-8 s)
		       (make-string 1 #\000)))
		    string-list))))

;; computes the offset for a given string within the #STRINGS file
(define (strings$offset s string-list)
  (let ((s (ignore-/ s)))  ; in case it's a filename
    (let loop ((lst string-list)
	       (offset 1))      ; the string at offset 0 is a null
      (cond
       ((null? lst) #f)
       ((or (not s) (string=? s "")) 0)
       ((string=? (car lst) s) offset)
       (else
	(loop (cdr lst) (+ offset 1 (string-length (car lst)))))))))

;; the #TOPICS "file"
(define (topics$ topics string-list)
  (apply
   string-append
   (filter-map
    (lambda (topic)
      (if (special? topic)
	  #f
	  (let* ((title (or (page-title (topic-data topic)) 
			    (ignore-/ (topic-name topic))))
		 (offset (strings$offset title string-list)))
	    (string-append
	     (dword 0)          ; Offset into the tree in the #TOCIDX file.
	     (dword offset)     ; title tag (#STRINGS offset)
	     (dword 0)          ; Offset in #URLTBL of entry containing offset to #URLSTR entry containing the URL
	     (word (if (in-contents? topic) 6 2))
	     (word 0)           ; unknown
	     ))))
    topics)))

;; the #URLSTR "file"
;; must be in 4000 byte blocks, except the last block
(define (urlstr$ topics)
  (string-append
   (apply
    string-append
    (filter-map
     (lambda (topic)
       (if (special? topic) ; a special topic like contents or index
	   #f
	   (string-append
	    (make-string 1 #\000)    ; unknown
	    (dword 0)           ; offset of URL
	    (dword 0)           ; offset of FrameName
	    (ignore-/ (topic-name topic))     ; local for this topic
	    )))
     topics))
   (make-string 1 #\000)   ; null terminator
   ))

;; the #URLTBL "file"
;; must be in broken up into 4096-byte blocks
(define (urltbl$)
  (let ((entries (string-append
		  (dword #xF8270122) ; unknown
		  (dword 0) ; Index of entry in #TOPICS file
		  (dword 1) ; Offset in #URLSTR file of entry containing filename
		  )))
    entries))
  
;; the NameList "file"
(define (namelist section-names)
  (let ((entry-list
	 (map (lambda (name)
		(string-append (word (string-length name))
			       (utf-16 name)
			       (word 0)))
	      section-names)))
    (string-append
     (word (+ 4 (byte-list-total entry-list)))  ; length of file
     (word (length entry-list))                 ; number of entries
     (apply string-append entry-list))))        ; the entries

;; lzx compression of an empty string:
(define (lzx-data)
  (string-append
   "\377\377\377\377\b\020\000\000\000\000\000\000\000\000"
   "\001\000\007\001\377\377\340\377\000\000\000\000\000\000"
   "\004\000\031B\276\317\357\373\371\276\000\340\000\000\000"
   "\000\000\000B\004\277\017\373\364\337\342\377\377\377\377"
   "\377\377\377\377\377\377\377\377\377\377\377\377\377\377"
   "\377\377\377\377\377\377\377\377\377\377\377\377\300\377"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; write out a chm file:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (main argv)
  (let ((out-path "out.chm")
	(in-dir ".")
	(default-topic "index.html")
	(title "Help File")
	(index "index.hhk")
	(contents "contents.hhc"))
    (args-parse (cdr argv)
      (section "Help")
      (("-o" ?file (help "Name of resulting chm file"))
       (set! out-path file))
      (("--output-file=?file" (help "Name of resulting chm file"))
       (set! out-path file))
      (("-i" ?dir (help "Directory containing input pages"))
       (set! in-dir dir))
      (("-d" ?default (help "Default topic filename."))
       (set! default-topic default))
      ((("-h" "--help") (help "This message."))
       (args-parse-usage #f)))
    (let* ((contents-topic (if (file-exists? (format "~a/~a" in-dir contents))
			       contents ""))
	   (index-topic (if (file-exists? (format "~a/~a" in-dir index))
			    index ""))
	   (topics (load-dir in-dir default-topic contents-topic index-topic))
	   (titles (map (lambda (t)
			  (let ((title (page-title (topic-data t))))
			    (if title title (ignore-/ (topic-name t)))))
			(keep-not special? topics)))
	   (string-list 
	    (keep-unique
	     (append (list title contents-topic index-topic)
		     (map (compose ignore-/ topic-name)
			  (keep-not special? topics))
		     titles)))
	   (namelist-data (namelist '("Uncompressed" "MSCompressed")))
	   (file-list
	     (append
	      (list
	       (make-topic "::DataSpace/NameList" 0 namelist-data #f)
	       (make-topic "/" 0 #f #f)
	       (make-topic "/#ITBITS" 0 #f #f)
	       (make-topic "/#SYSTEM" 0 (system$ topics title) #f)
	       (make-topic "/#IDXHDR" 0 (idxhdr$ topics) #f)
	       (make-topic "/#STRINGS" 0 (strings$ string-list) #f)
	       (make-topic "/#TOPICS" 0 (topics$ topics string-list) #f)
	       (make-topic "/#URLSTR" 0 (urlstr$ topics) #f)
	       (make-topic "/#URLTBL" 0 (urltbl$) #f)
	       (make-topic "/$FIftiMain" 0 #f #f)
	       (make-topic "/#IVB" 0 (dword 0) #f)  ; total size is 0
	       (make-topic "/$OBJINST" 0 #f #f) ; not neede d
	       (make-topic "/$WWAssociativeLinks/" 0 #f #f)
	       (make-topic "/$WWAssociativeLinks/Property" 0 (dword 0) #f)
	       (make-topic "/$WWKeywordLinks/" 0 #f #f)
	       (make-topic "/$WWKeywordLinks/Property" 0 (dword 0) #f)
	       )
	      topics))
	   (sorted-list (sort file-list (lambda (a b) (string-ci<? (topic-name a) (topic-name b)))))
	   (content-listing
	    (let loop ((lst (append
			     sorted-list
			     ;; we tack on an empty compressed section:
			     (list (make-topic "::DataSpace/Storage/MSCompressed/Contents" 1 "" #f))))
		       (offset 0)
		       (acc '()))
	      (if (null? lst)
		  (reverse acc)
		  (let* ((entry (car lst))
			 (data (topic-data entry))
			 (len (if data (string-length data) 0)))
		    (loop (cdr lst)
			  (+ offset len)
			  (cons (list (topic-name entry)
				      (topic-section entry)
				      data
				      (if data offset 0)
				      len)
				acc))))))
	   (cont-0-data (apply string-append
			       (remove-false
				(map topic-data sorted-list))))
	   (cont-1-data (lzx-data))
	   (dir-listings (listing-chunks content-listing))
	   (index-chunk-list '())
	   (filesize (+ (string-length (header index-chunk-list dir-listings 0))
			(string-length cont-0-data) (string-length cont-1-data))))
      ;;(printf "content-listing: \n") (pp content-listing)
      (with-output-to-file out-path
	(lambda ()
	  (display (header index-chunk-list dir-listings filesize))
	  (display cont-0-data)
	  (display cont-1-data))))))
